

from typing import Iterator, Optional, TypeVar, Tuple, List

import torch, os, abc
import numpy as np
import torchdata.datapipes as dp
from functools import lru_cache
from freeplot.utils import import_pickle, export_pickle

from ..tags import FieldTags, SPARSE, USER, SESSION, ITEM, ID
from ..fields import Field, BufferField, FieldList, FieldTuple
from ..utils import download_from_url, extract_archive
from ...utils import timemeter, infoLogger, mkdirs, warnLogger


__all__ = ['BaseSet', 'RecDataSet']


DEFAULT_PICKLE_FMT = "{0}_from_pickle"
DEFAULT_TRANSFORM_FILENAME = "transforms.pickle"
DEFAULT_CHUNK_FMT = "chunk{0}.pickle"


T = TypeVar('T')


#===============================Basic Class===============================

class RecSetBuildingError(Exception): ...


class BaseSet(dp.iter.IterDataPipe, metaclass=abc.ABCMeta):
    r""" 
    Base class for data pipes. Defines basic functionality and methods for 
    pre-processing the data for the learning models.

    Parameters:
    -----------
    root: str
        The path storing datasets.
    filename: str, optional 
        The dirname of the dataset. If `None`, sets the classname as the filename.
    download: bool 
        Download the dataset from a URL.

    Attributes:
    -----------
    _cfg: Config[str, Field] 
        Includes fields of each column.
    DEFAULT_CHUNK_SIZE: int, default 51200 
        Chunk size for saving.
    DATATYPE: str
        Dataset type.
        - `General': for general recommendation.
        - `Sequential': for sequential recommendation.
        - `Session': for session-based recommendation.
        - `Context': for context-aware recommendation.
        - `Knowledge': for knowledge-based recommendation.
    VALID_IS_TEST: bool 
        The validset and testset are the same one sometimes.

    Notes:
    ------
    All datasets that inherit RecDataSet should define the class variable `_cfg` before instantiation.
    Generally speaking, the dataset will be split into:
        - trainset
        - validset
        - testset
    """

    DEFAULT_CHUNK_SIZE = 51200 # chunk size
    URL: str
    DATATYPE: str
    VALID_IS_TEST: bool

    def __new__(cls, *args, **kwargs):
        for attr in ('_cfg', 'DATATYPE', 'VALID_IS_TEST'):
            if not hasattr(cls, attr):
                raise RecSetBuildingError(f"'{attr}' should be defined before instantiation ...")
        assert 'fields' in cls._cfg, "the config of fields should be defined in '_cfg' ..."
        return super().__new__(cls)

    def __init__(self, root: str, filename: Optional[str] = None, download: bool = True) -> None:
        super().__init__()

        self.trainsize: int = 0
        self.validsize: int = 0
        self.testsize: int = 0

        fields = []
        for field_type, cfg in self._cfg['fields']:
            fields.append(field_type(**cfg))
        self.fields = fields
        self.__mode = 'train'

        filename = filename if filename else self.__class__.__name__
        self.path = os.path.join(root, self.DATATYPE, filename)
        if not os.path.exists(self.path) or not any(True for _ in os.scandir(self.path)):
            if download:
                extract_archive(
                    download_from_url(self.URL, root, overwrite=False),
                    self.path
                )
            else:
                raise FileNotFoundError(f"No such file of {self.path}, or this dir is empty ...")

        self.compile()
        self.check()

    def check(self):
        """Self-check program should be placed here."""
        ...
        
    @property
    def mode(self) -> str:
        r"""
        Return the mode in which the dataset is currently being used.

        Returns:
            str: The mode in which the dataset is currently being used.
        """

        return self.__mode

    def train(self: T) -> T:
        """Switch the dataset mode to 'train'."""
        self.__mode = 'train'
        return self

    def valid(self: T) -> T:
        """Switch the dataset mode to 'valid'."""
        self.__mode = 'valid'
        return self

    def test(self: T) -> T:
        """Switch the dataset mode to 'test'."""
        self.__mode = 'test'
        return self

    @property
    def fields(self):
        return self.__fields

    @fields.setter
    def fields(self, vals) -> FieldTuple[Field]:
        """Return: Tuple of Field."""
        self.__fields = FieldTuple(vals)

    @abc.abstractmethod
    def raw2data(self) -> dp.iter.IterableWrapper:
        r"""
        Process raw data.

        Returns:
        --------
        A processed data.

        Raises:
        -------
        NotImplementedError: Subclasses should implement this method.

        Notes:
        ------
        This method should be implemented by subclasses.
        """
        raise NotImplementedError(f"{self.__class__.__name__}.raw2data() method should be implemented ...")

    @abc.abstractmethod
    def pickle2data(self):
        raise NotImplementedError(
                f"{self.__class__.__name__}.pickle2data should be implemented before using ..."
        )

    def __iter__(self) -> Iterator[FieldList[BufferField]]:
        for cols in self.pickle2data():
            yield FieldList(map(
                lambda field, col: field.buffer(col),
                self.fields,
                cols
            ))

    def summary(self):
        """Print a summary of the dataset."""
        infoLogger(str(self))

    @timemeter
    def to_heterograph(self, *edge_types: Tuple[Tuple[FieldTags], Optional[str], Tuple[FieldTags]]):
        r"""
        Convert datapipe to a heterograph.

        Parameters:
        -----------
        *edge_types: Tuple[Tuple[str, str], Optional[str], Tuple[str, str]]
            The desired edges in the returned heterograph. Each edge is defined by a tuple (source, edge, destination):
            - source: Tuple of field tags for filtering the source node data.
            - edge: Optional[str], default None
                The name of the edge. If not provided, the name 'src.field2dst.field' will be used.
            - destination: Tuple of field tags for filtering the destination node data.

        Returns:
        --------
        HeteroData:
            The resulting heterograph, with the requested edges and nodes.

        Notes:
        ------
        A warning will be raised if the current mode is not 'train'!

        Examples:
        ---------
        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_heterograph(
        ...    ((USER, ID), None, (ITEM, ID)),
        ...    ((ITEM, ID), None, (USER, ID))
        ... )
        >>> graph
        HeteroData(
            UserID={ x=[29858, 0] },
            ItemID={ x=[40981, 0] },
            (UserID, UserID2ItemID, ItemID)={ edge_index=[2, 810128] },
            (ItemID, ItemID2UserID, UserID)={ edge_index=[2, 810128] }
        )
        """
        from torch_geometric.data import HeteroData

        # check mode and raise warning if not in 'train' mode
        if self.mode != 'train':
            warnLogger(f"Convert the datapipe for {self.mode} to graph. Make sure that this is intentional ...")

        srcs, edges, dsts = zip(*edge_types)
        srcs = [self.fields[src] for src in srcs]
        dsts = [self.fields[dst] for dst in dsts]
        nodes = set(srcs + dsts)
        edges = list(map(
            lambda src, edge, dst: edge if edge else f"{src.name}2{dst.name}",
            srcs, edges, dsts
        ))
        data = {node.name: [] for node in nodes}
        for chunk in self:
            for node in nodes:
                data[node.name].append(np.ravel(chunk[node.tags].data))
        for key in data:
            data[key] = torch.tensor(np.concatenate(data[key], axis=0), dtype=torch.long)

        graph = HeteroData()
        for node in srcs:
            graph[node.name].x = torch.empty((node.count, 0), dtype=torch.long)
        for node in dsts:
            if node not in srcs:
                graph[node.name].x = torch.empty((node.count, 0), dtype=torch.long)
        for src, edge, dst in zip(srcs, edges, dsts):
            u, v = data[src.name], data[dst.name]
            graph[src.name, edge, dst.name].edge_index = torch.stack((u, v), dim=0) # 2 x N
        return graph.coalesce()

    def to_bigraph(
        self,
        src: Tuple[FieldTags] = (USER, ID),
        dst: Tuple[FieldTags] = (ITEM, ID),
        edge_type: Optional[str] = None
    ):
        r"""
        Convert datapipe to a bipartite graph.

        Parameters:
        ----------
        src: Tuple[FieldTags] 
            Source node.
        dst: Tuple[FieldTags] 
            Destination node.
        edge_type: str, optional 
            The name of the edge. `src.name2dst.name` will be specified if `edge_type` is `None`.

        Returns:
        --------
        HeteroData:
            The resulting heterograph, with the requested edges and nodes.

        Notes:
        ------
        A warning will be raised if the current mode is not 'train'!

        Examples:
        ---------
        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_bigraph(
        ...    (USER, ID), (ITEM, ID)
        ... )
        >>> graph
        HeteroData(
            UserID={ x=[29858, 0] },
            ItemID={ x=[40981, 0] },
            (UserID, UserID2ItemID, ItemID)={ edge_index=[2, 810128] }
        )
        """
        return self.to_heterograph((src, edge_type, dst))
   
    def to_graph(self, src: Tuple[FieldTags], dst: Tuple[FieldTags]):
        r"""
        Convert datapipe to a homogeneous graph.

        Parameters:
        ----------
        src: Tuple[FieldTags]
            Source node.
        dst: Tuple[FieldTags]
            Destination node.

        Returns:
        --------
        HeteroData:
            The resulting heterograph, with the requested edges and nodes.

        Notes:
        ------
        A warning will be raised if current mode is not 'train' !

        Examples:
        --------
        >>> from freerec.data.datasets import Gowalla_m1
        >>> from freerec.data.tags import USER, ITEM, ID
        >>> basepipe = Gowalla_m1("../data")
        >>> fields = basepipe.fields
        >>> graph = basepipe.to_graph(
        ...    fields[USER, ID], fields[ITEM, ID],
        ... )
        >>> graph
        Data(edge_index=[2, 1620256], x=[70839, 0])
        """
        from torch_geometric.utils import to_undirected
        graph = self.to_heterograph((src, None, dst)).to_homogeneous()
        graph.edge_index = to_undirected(graph.edge_index)
        return graph


class RecDataSet(BaseSet):
    """RecDataSet provides a template for specific datasets."""

    def check_transforms(self) -> None:
        """Check if the transformations exist."""
        file_ = os.path.join(
            self.path,
            DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            DEFAULT_TRANSFORM_FILENAME
        )
        if os.path.isfile(file_):
            return True
        else:
            mkdirs(os.path.join(
                self.path,
                DEFAULT_PICKLE_FMT.format(self.__class__.__name__)
            ))
            return False

    def save_transforms(self) -> None:
        """Save transformers in a pickle format."""
        infoLogger(f"[{self.__class__.__name__}] >>> Save transformers ...")
        state_dict = self.fields.state_dict()
        state_dict['trainsize'] = self.trainsize
        state_dict['validsize'] = self.validsize
        state_dict['testsize'] = self.testsize
        file_ = os.path.join(
            self.path, 
            DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            DEFAULT_TRANSFORM_FILENAME
        )
        export_pickle(state_dict, file_)

    def load_transforms(self) -> None:
        """Load transformers from a pickle file."""
        file_ = os.path.join(
            self.path, 
            DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            DEFAULT_TRANSFORM_FILENAME
        )
        state_dict = import_pickle(file_)
        self.trainsize = state_dict['trainsize']
        self.validsize = state_dict['validsize']
        self.testsize = state_dict['testsize']
        self.fields.load_state_dict(state_dict, strict=False)

    def check_pickle(self) -> bool:
        """Check if the dataset has been converted into pickle format."""
        path = os.path.join(
            self.path,
            DEFAULT_PICKLE_FMT.format(self.__class__.__name__), 
            self.mode
        )
        if os.path.exists(path):
            return any(True for _ in os.scandir(path))
        else:
            os.makedirs(path)
            return False

    def write_pickle(self, data, count: int) -> None:
        r"""
        Save pickle format data.

        Parameters:
        -----------
        data: Any
            The data to be saved in pickle format.
        count: int
            The count of the data chunks.

        Returns:
        --------
        None
        """
        file_ = os.path.join(
            self.path,
            DEFAULT_PICKLE_FMT.format(self.__class__.__name__),
            self.mode, DEFAULT_CHUNK_FMT.format(count)
        )
        export_pickle(data, file_)

    def read_pickle(self, file_: str):
        r"""
        Load pickle format data.

        Parameters:
        -----------
        file_: str 
            The file path to load the pickle format data.

        Returns:
        --------
        The loaded pickle format data.
        """
        return import_pickle(file_)

    def row_processer(self, row):
        r"""
        Process a row of raw data.

        Parameters:
        -----------
        row: Any
            A row of raw data.

        Returns:
        --------
        A processed row of data.
        """
        return [field.caster(val) for val, field in zip(row, self.fields)]

    def raw2pickle(self):
        """Convert raw data into pickle format."""
        infoLogger(f"[{self.__class__.__name__}] >>> Convert raw data ({self.mode}) to chunks in pickle format")
        datapipe = self.raw2data()
        count = 0
        for chunk in datapipe.batch(batch_size=self.DEFAULT_CHUNK_SIZE).column_():
            for j, field in enumerate(self.fields):
                chunk[j] = field.transform(chunk[j])
            self.write_pickle(chunk, count)
            count += 1
        infoLogger(f"[{self.__class__.__name__}] >>> {count} chunks done")

    def pickle2data(self):
        r"""
        Read the pickle data and return it as a generator.

        Yields:
        -------
        A chunk of the pickle data.
        """
        datapipe = dp.iter.FileLister(
            os.path.join(
                self.path,
                DEFAULT_PICKLE_FMT.format(self.__class__.__name__),
                self.mode
            ),
            non_deterministic=False # return sorted chunks
        )
        for file_ in datapipe:
            yield self.read_pickle(file_)

    @timemeter
    def compile(self):
        r"""
        Check current dataset and transformations.

        Flows:
        ------
        1. Check whether the transformation has been fitted:
            - `True`: Skip.
            - `False`: Fit the total trainset and the `SPARSE` fields in valid|testset
                to avoid unseen features. This operation will not cause information leakage.
            
        2. Convert each set into pickle format for fast loading.
        """

        def fit_transform(*tags):
            """Function to fit and transform data for pickle conversion."""
            datapipe = self.raw2data().batch(self.DEFAULT_CHUNK_SIZE).column_()
            datasize = 0
            fields = self.fields.groupby(*tags)
            for chunk in datapipe:
                datasize += len(chunk[0])
                for field in fields:
                    index = self.fields.index(*field.tags)
                    field.partial_fit(chunk[index])
            return datasize

        if self.check_transforms():
            self.load_transforms()
        else:
            self.train()
            self.trainsize = fit_transform()

            # avoid unseen IDs not included in trainset
            self.valid()
            self.validsize = fit_transform(SPARSE)
            self.test()
            self.testsize = fit_transform(SPARSE)

            self.save_transforms()
            
        # raw2pickle
        self.train()
        if not self.check_pickle():
            self.raw2pickle()
        self.valid()
        if not self.check_pickle():
            self.raw2pickle()
        self.test()
        if not self.check_pickle():
            self.raw2pickle()
        self.train()

        self.summary()

    @property
    def datasize(self):
        """Return the size of dataset according to the current mode."""
        if self.mode == 'train':
            return self.trainsize
        elif self.mode == 'valid':
            return self.validsize
        else:
            return self.testsize

    def to_pairs(self, master: Tuple = (USER, ID)) -> List:
        r"""
        Return dataset in pairs.

        Parameters:
        -----------
        master: Tuple
            Tuple of tags to spefic a field, e.g., (USER, ID), (SESSION, ID)
        
        Returns:
        --------
        List
        """
        Master = self.fields[master]
        assert Master is not None, f"{Master} is not in fields ..."
        pairs = []

        for chunk in self:
            pairs.extend(list(zip(chunk[master], chunk[ITEM, ID])))
        return pairs

    def to_seqs(self, master: Tuple = (USER, ID), keepid: bool = False) -> List:
        r"""
        Return dataset in sequence.

        Parameters:
        -----------
        master: Tuple
            Tuple of tags to spefic a field, e.g., (USER, ID), (SESSION, ID)
        keepid: bool, default to False
            `True`: return list of (id, items)
            `False`: return list of items
        
        Returns:
        --------
        List
        """
        Master = self.fields[master]
        assert Master is not None, f"{Master} is not in fields ..."
        seqs = [[] for id_ in range(Master.count)]

        for chunk in self:
            list(map(
                lambda id_, item: seqs[id_].append(item),
                chunk[master], chunk[ITEM, ID]
            ))

        if keepid:
            seqs = [(id_, tuple(items)) for id_, items in enumerate(seqs)]
        else:
            seqs = [tuple(items) for items in seqs]

        return seqs

    def to_roll_seqs(
        self, master: Tuple = (USER, ID), 
        minlen: int = 2,
        maxlen: Optional[int] = None
    ) -> List:
        r"""
        Rolling dataset in sequence.

        Parameters:
        -----------
        master: Tuple
            Tuple of tags to spefic a field, e.g., (USER, ID), (SESSION, ID)
        minlen: int
            Shorest sequence
        maxlen: int, optional
            Maximum length
            `None`: Roll throughout the whole sequence
       
        Returns:
        --------
        List
        """
        seqs = self.to_seqs(master=master, keepid=True)

        roll_seqs = []
        for id_, items in seqs:
            if maxlen is not None:
                items = items[-maxlen:]
            for k in range(minlen, len(items) + 1):
                roll_seqs.append(
                    (id_, items[:k])
                )

        return roll_seqs

    def seqlens(self, master: Tuple = (USER, ID)) -> List:
        seqs = self.to_seqs(master, keepid=False)
        return list(filter(lambda x: x > 0, [len(items) for items in seqs]))

    @lru_cache()
    def has_duplicates(self, master: Tuple = (USER, ID)) -> bool:
        r"""
        Check whether the dataset has repeated interactions.

        Parameters:
        -----------
        master: Tuple
            Tuple of tags to spefic a field, e.g., (USER, ID), (SESSION, ID)

        Returns:
        --------
        bool
        """
        from itertools import chain
        train_seqs = self.train().to_seqs(master, keepid=False)
        valid_seqs = self.valid().to_seqs(master, keepid=False)
        test_seqs = self.test().to_seqs(master, keepid=False)
        seqs = map(
            lambda triple: chain(*triple),
            zip(train_seqs, valid_seqs, test_seqs)
        )
        for seq in seqs:
            seq = list(seq)
            if len(seq) != len(set(seq)):
                return True
        return False

    @property
    def maxlen(self) -> int:
        return np.max(self.seqlens()).item()

    @property
    def minlen(self) -> int:
        return np.min(self.seqlens()).item()

    @property
    def meanlen(self) -> int:
        return np.mean(self.seqlens()).item()

    def __str__(self) -> str:
        cfg = '\n'.join(map(str, self.fields))
        return f"[{self.__class__.__name__}] >>> \n" + cfg



import setuptools
import re

with open("README.md", "r") as fh:
  long_description = fh.read()

requires = [
    # "torchdata==0.4.1",
    "freeplot>=0.4.5",
    "PyYAML>=6.0",
    "tensorboard>=2.10.0",
    "psutil>=5.9.0",
    "prettytable>=3.4.1"
]


def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open(project + '/__init__.py').read())
    return result.group(1)


setuptools.setup(
  name="freerec",
  entry_points={
    'console_scripts': [
      'freerec=freerec.__main__:main',
    ],
  },
  version=get_property('__version__', 'freerec'),
  author="MTandHJ",
  author_email="congxueric@gmail.com",
  description="PyTorch library for recommender systems",
  long_description=long_description,
  long_description_content_type="text/markdown",
  license='MIT License',
  url="https://github.com/MTandHJ/freerec",
  packages=setuptools.find_packages(),
  python_requires='>=3.9',
  install_requires=requires,
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)




from typing import Iterable, Union, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F


__all__ = [
    "BaseCriterion", 
    "CrossEntropy4Logits", "KLDivLoss4Logits",
    "BPRLoss", "BCELoss4Logits",
    "MSELoss", "L1Loss",
]


class BaseCriterion(nn.Module):
    r"""
    Criterion.

    Parameters:
    -----------
    reduction : str, default 'mean'
        Reduction method. Choices are 'none', 'sum', and 'mean'.

    Attributes:
    -----------
    reduction : str
        Reduction method.

    Methods:
    --------
    regularize(params: Union[torch.Tensor, Iterable[torch.Tensor]], rtype: str = 'l2') -> torch.Tensor
        Regularizes the given parameters with the specified regularization method.

    """

    def __init__(self, reduction: str = 'mean') -> None:
        super().__init__()
        self.reduction = reduction

    @staticmethod
    def regularize(params: Union[torch.Tensor, Iterable[torch.Tensor]], rtype: str = 'l2'):
        r"""
        Add regularization for given parameters.

        Parameters:
        -----------
        params : Union[torch.Tensor, Iterable[torch.Tensor]]
            List of parameters for regularization.
        rtype : str, default 'l2'
            The type of regularization to use. Options include 'l1' and 'l2'.

        Returns:
        --------
        reg_loss : torch.Tensor
            The regularization loss tensor.
        """
        params = [params] if isinstance(params, torch.Tensor) else params
        if rtype == 'l1':
            return sum(param.abs().sum() for param in params)
        elif rtype == 'l2':
            return sum(param.pow(2).sum() for param in params) / 2
        else:
            raise NotImplementedError(f"{rtype} regularization is not supported ...")


class CrossEntropy4Logits(BaseCriterion):
    """Cross entropy loss with logits."""

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor, 
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return cross_entropy_with_logits(logits, targets, reduction)


class BCELoss4Logits(BaseCriterion):
    """Binary Cross Entropy with logits !!!"""

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor,
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return binary_cross_entropy_with_logits(logits, targets, reduction)


class KLDivLoss4Logits(BaseCriterion):
    """KLDivLoss with logits"""

    def __init__(self, reduction: str = 'batchmean') -> None:
        super().__init__(reduction)

    def forward(
        self, logits: torch.Tensor, targets: torch.Tensor,
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return kl_div_loss_with_logits(logits, targets, reduction)
        

class BPRLoss(BaseCriterion):

    def forward(
        self, pos_scores: torch.Tensor, neg_scores: torch.Tensor,
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return bpr_loss_with_logits(pos_scores, neg_scores, reduction)


class MSELoss(BaseCriterion):
    """Mean Square Loss."""

    def forward(
        self, inputs: torch.Tensor, targets: torch.Tensor,
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return F.mse_loss(inputs, targets, reduction=reduction)


class L1Loss(BaseCriterion):
    """L1 Loss."""

    def forward(
        self, inputs: torch.Tensor, targets: torch.Tensor,
        reduction: Optional[str] = None
    ):
        reduction = reduction if reduction is not None else self.reduction
        return F.l1_loss(inputs, targets, reduction=reduction)





def cross_entropy_with_logits(
    logits: torch.Tensor, targets: torch.Tensor,
    reduction: str = "mean"
) -> torch.Tensor:
    r"""
    Cross-entropy loss with logits.

    Parameters:
    -----------
    logits: torch.Tensor
    targets: torch.Tensor
    reduction: str, 'none'|'sum'|'mean' (default)

    Shapes:
    -------
    logits:     (C,)    (B, C)  (B, C, ...)
    targets:    ( ,)    (C,  )  (B, ...)
    """
    return F.cross_entropy(logits, targets, reduction=reduction)

def binary_cross_entropy_with_logits(
    logits: torch.Tensor, targets: torch.Tensor,
    reduction: str = "mean"
) -> torch.Tensor:
    r"""
    Binary cross-entropy loss with logits.

    Parameters:
    -----------
    logits: torch.Tensor
    targets: torch.Tensor, the same shape as `logits`
    reduction: str, 'none'|'sum'|'mean' (default)
    """
    return F.binary_cross_entropy_with_logits(logits, targets.to(logits.dtype), reduction=reduction)

def kl_div_loss_with_logits(
    logits: torch.Tensor, targets: torch.Tensor,
    reduction: str = "batchmean"
) -> torch.Tensor:
    r"""
    KL divergence loss with logits.

    Parameters:
    -----------
    logits: torch.Tensor
    targets: torch.Tensor, the same shape as `logits`
    reduction: str, 'none'|'sum'|'mean'|'batchmean' (default)
    """
    assert logits.size() == targets.size()
    inputs = F.log_softmax(logits, dim=-1)
    targets = F.softmax(targets, dim=-1)
    return F.kl_div(inputs, targets, reduction=reduction)

def bpr_loss_with_logits(
    pos_scores: torch.Tensor, neg_scores: torch.Tensor,
    reduction: str = "mean"
) -> torch.Tensor:
    r"""
    BPR loss with logits.

    Parameters:
    -----------
    pos_scores: torch.Tensor
        Positive scores
    neg_scores: torch.Tensor
        Negative scores
    reduction: str, 'none'|'sum'|'mean' (default)
    """
    loss = F.softplus(neg_scores - pos_scores)
    if reduction == 'none':
        return loss
    elif reduction == 'mean':
        return loss.mean()
    elif reduction == 'sum':
        return loss.sum()
    else:
        raise NotImplementedError(f"reduction mode of '{reduction}' is not supported ...")




from typing import TypeVar, Callable, Optional, Union, List, Tuple

import numpy as np
import os, requests, warnings, hashlib, tqdm

from ..utils import infoLogger


T = TypeVar('T')
class DataSetLoadingError(Exception): ...

def safe_cast(val: T, dest_type: Callable[[T], T], default: T) -> T:
    r"""
    Cast the value to the specified type.

    Parameters:
    -----------
    val : T
        The value to be casted to the specified type.
    dest_type : Callable[[T], T]
        A function to cast `val` to the specified type.
    default : T
        The default value to use if `val` is None or an empty string.

    Returns:
    --------
    value: T
        The value of `val` casted to the specified type.

    Raises:
    -------
    ValueError
        If `val` or `default` cannot be casted to the specified type `dest_type`.

    Notes:
    ------
    This function casts `val` to the specified type using the `dest_type` function.
    If `val` is None or an empty string, the function will use the `default` value.
    If the `default` value is None, a ValueError will be raised.
    """
    try:
        if val:
            return dest_type(val)
        else: # fill_na
            if default is None:
                raise ValueError
            return dest_type(default)
    except (ValueError, TypeError):
        raise ValueError(
            f"Using '{dest_type.__name__}' to convert '{val}' where the default value is {default} ..." \
            f"This happens when the value (or the default value: {default}) to be cast is not of the type {dest_type.__name__}.",
        )


def download_from_url(
    url: str, root: str = '.', filename: Optional[str] = None, 
    overwrite: bool = False, retries=5, 
    sha1_hash: Optional[str] = None, verify_ssl: bool = True, 
    log: bool = True
):
    r"""
    Download a file from a given URL.

    Codes borrowed from dgl.data.utils

    Parameters:
    -----------
    url : str
        The URL to download the file from.
    root : str, optional
        The root directory where the downloaded file will be saved. Default is current directory ('.').
    filename: str, optional
        The filename of the downloaded file. If None, the filename will be inferred from the URL.
    overwrite : bool, optional
        Whether to overwrite the destination file if it already exists.
    sha1_hash : str, optional
        Expected sha1 hash in hexadecimal digits. Will ignore existing file when hash is specified
        but doesn't match.
    retries : int, optional
        The number of times to attempt downloading in case of failure or non 200 return codes. Default is 5.
    verify_ssl : bool, optional
        Verify SSL certificates. Default is True.
    log : bool, optional
        Whether to print the progress of download. Default is True.

    Returns:
    --------
    str
        The file path of the downloaded file.
    """
    if filename is None:
        filename = url.split('/')[-1]
        # Empty filenames are invalid
        assert filename, 'Can\'t construct file-name from this URL. ' \
            'Please set the `path` option manually.'
    file_ = os.path.join(root, filename)

    assert retries >= 0, "Number of retries should be at least 0"

    if not verify_ssl:
        warnings.warn(
            'Unverified HTTPS request is being made (verify_ssl=False). '
            'Adding certificate verification is strongly advised.')

    if overwrite or not os.path.exists(file_) or (sha1_hash and not check_sha1(file_, sha1_hash)):
        dirname = os.path.dirname(os.path.abspath(os.path.expanduser(file_)))
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        while retries+1 > 0:
            # Disable pyling too broad Exception
            # pylint: disable=W0703
            try:
                if log:
                    infoLogger('[DataSet] >>> Downloading %s from %s...' % (file_, url))
                r = requests.get(url, stream=True, verify=verify_ssl)
                if r.status_code != 200:
                    raise RuntimeError("Failed downloading url %s" % url)
                with open(file_, 'wb') as f:
                    for chunk in tqdm.tqdm(r.iter_content(chunk_size=1024), leave=False, desc="վ'ᴗ' ի-"):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                if sha1_hash and not check_sha1(file_, sha1_hash):
                    raise DataSetLoadingError(
                        'File {} is downloaded but the content hash does not match.'
                        ' The repo may be outdated or download may be incomplete. '
                        'If the "repo_url" is overridden, consider switching to '
                        'the default repo.'.format(file_)
                    )
                break
            except Exception as e:
                retries -= 1
                if retries <= 0:
                    raise e
                else:
                    if log:
                        infoLogger("[DataSet] >>> Download failed, retrying, {} attempt{} left"
                              .format(retries, 's' if retries > 1 else ''))

    return file_


def extract_archive(file_, target_dir, overwrite=False):
    r"""
    Extract files from an archive.

    Codes borrowed from dgl/data/utils.py

    Parameters:
    -----------
    file_ : str
        The path to the archive file to extract.
    target_dir : str
        The directory to extract the archive contents to.
    overwrite : bool, optional
        Whether to overwrite existing files in the target directory. Defaults to False.

    Raises:
    -------
    DataSetLoadingError
        If an unsupported archive file type is encountered.
    """

    if os.path.exists(target_dir) and not overwrite:
        return
    infoLogger('[DataSet] >>> Extracting file to {}'.format(target_dir))
    if file_.endswith('.tar.gz') or file_.endswith('.tar') or file_.endswith('.tgz'):
        import tarfile
        with tarfile.open(file_, 'r') as archive:
            archive.extractall(path=target_dir)
    elif file_.endswith('.gz'):
        import gzip
        import shutil
        with gzip.open(file_, 'rb') as f_in:
            target_file = os.path.join(target_dir, os.path.basename(file_)[:-3])
            with open(target_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    elif file_.endswith('.zip'):
        import zipfile
        with zipfile.ZipFile(file_, 'r') as archive:
            archive.extractall(path=target_dir)
    else:
        raise DataSetLoadingError('Unrecognized file type: ' + file_)


def check_sha1(filename, sha1_hash):
    r"""
    Check if the SHA1 hash of a file matches the expected hash.

    Parameters:
    ----------
    filename : str
        The path to the file to check the hash of.
    sha1_hash : str
        The expected SHA1 hash value.

    Returns:
    --------
    bool
        True if the file's hash matches the expected value, False otherwise.
    """
    sha1 = hashlib.sha1()
    with open(filename, 'rb') as f:
        while True:
            data = f.read(1048576)
            if not data:
                break
            sha1.update(data)

    return sha1.hexdigest() == sha1_hash


def negsamp_vectorized_bsearch(
    positives: Union[Tuple, List], n_items: int, 
    size: Union[int, List[int], Tuple[int]] = 1,
    replacement: bool = True
) -> np.ndarray:
    r"""
    Uniformly sampling negatives according to a list of ordered positives
    See [here](https://tech.hbc.com/2018-03-23-negative-sampling-in-numpy.html) for more details.

    Parameters:
    -----------
    positives: Union[Tuple, List], 1-D
        The positive indices should be ordered.
    n_items: int
        The number of all items.
    size: int or List[int] or Tuple[int]
        The size of required negatives.
    replacement: bool, default to `True`
        - True: sampling with replacement
        - False: sampling without replacement
    
    Raises:
    -------
    AssertionError:
        Given `positives` is not 1-D.
    ValueError:
        Too much negatives required.
    
    Returns:
    --------
    neg_inds: np.ndarray
        A list of negatives in the size of `size`.
    """
    positives = np.array(positives, copy=False)
    assert positives.ndim == 1, f"positives should be 1-D array but {positives.ndim}-D received ..."
    try:
        if replacement:
            raw_samp = np.random.randint(0, n_items - len(positives), size=size)
        else:
            raw_samp = np.random.choice(n_items - len(positives), size=size, replace=replacement)
    except ValueError:
        raise ValueError(
            "The number of required negatives is larger than that of candidates, but replacement is False ..."
        )
    pos_inds_adj = positives - np.arange(len(positives))
    ss = np.searchsorted(pos_inds_adj, raw_samp, side='right')
    neg_inds = raw_samp + ss
    return neg_inds




from typing import Optional, Union, overload, Tuple, Dict

import torch, inspect
import torch.nn as nn


__all__ = ['RecSysArch']


class RecSysArch(nn.Module):
    r"""
    A PyTorch Module for recommendation system architecture.
    This module contains methods for broadcasting tensors and initializing the module parameters.

    Parameters:
    -----------
    nn.Module : PyTorch Module
        The base class of this module.

    Methods:
    --------
    to(device, dtype, non_blocking)
        Moves and/or casts the parameters and buffers.

    initialize()
        Initializes the module parameters.

    broadcast(*tensors)
        Broadcasts the given tensors according to broadcasting semantics.

    """

    def __init__(self) -> None:
        super().__init__()
        self._check()

    @classmethod
    def _check(cls):
        signature = inspect.signature(cls.recommend_from_pool)
        if 'pool' not in signature.parameters:
            raise NotImplementedError(
                f"`pool` must be a argument of `recommend_from_pool` ..."
            )

    def to(
        self, device: Optional[Union[int, torch.device]] = None,
        dtype: Optional[Union[torch.dtype, str]] = None,
        non_blocking: bool = False
    ):
        r"""
        Moves and/or casts the parameters and buffers.

        Parameters:
        -----------
        device : Union[int, torch.device], optional
            The destination device of the parameters and buffers, by default None.
        dtype : Union[torch.dtype, str], optional
            The desired data type of the parameters and buffers, by default None.
        non_blocking : bool, optional
            Whether the copy should be asynchronous or not, by default False.

        Returns:
        --------
        nn.Module
            The module with parameters and buffers moved and/or cast.
        """
        if device:
            self.device = device
        return super().to(device, dtype, non_blocking)

    def reset_parameters(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.)
            elif isinstance(m, nn.Embedding):
                nn.init.xavier_normal_(m.weight)
            elif isinstance(m, nn.GRU):
                nn.init.xavier_uniform_(m.weight_hh_l0)
                nn.init.xavier_uniform_(m.weight_ih_l0)
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d)):
                nn.init.constant_(m.weight, 1.)
                nn.init.constant_(m.bias, 0.)

    @staticmethod
    def broadcast(*tensors: torch.Tensor):
        r"""
        Broadcasts the given tensors according to Broadcasting semantics.
        See [here](https://pytorch.org/docs/stable/generated/torch.broadcast_tensors.html#torch.broadcast_tensors) for details.

        Parameters:
        -----------
        tensors : torch.Tensor
            The tensors to broadcast.

        Returns:
        --------
        Tuple[torch.Tensor, ...]
            The broadcasted tensors.

        Examples:
        ---------
        >>> users = torch.rand(4, 1, 4)
        >>> items = torch.rand(4, 2, 1)
        >>> users, items = RecSysArch.broadcast(users, items)
        >>> users.size()
        torch.Size(4, 2, 4)
        >>> items.size()
        torch.Size(4, 2, 4)
        """
        return torch.broadcast_tensors(*tensors)

    @overload
    def recommend_from_full(self) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""
        Returns:
        --------
        user features: torch.Tensor
        item features: torch.Tensor
        """

    @overload
    def recommend_from_full(self, seqs: torch.Tensor) -> torch.Tensor:
        r"""
        Full ranking recommendation.

        Parameters:
        -----------
        seqs: torch.Tensor, (B, S)
            A batch of sequences.
        
        Returns:
        --------
        scores: torch.Tensor, (B, N)
            `N' denotes the number of items.
        """

    @overload
    def recommend_from_pool(self, seqs: torch.Tensor, pool: torch.Tensor) -> torch.Tensor:
        r"""
        Sampling-based recommendation.

        Parameters:
        -----------
        seqs: torch.Tensor, (B, S)
            A batch of sequences.
        pool: torch.Tensor, (B, K)
            A batch of items.

        Returns:
        --------
        scores: torch.Tensor, (B, K)
        """

    def recommend_from_full(self, **kwargs):
        raise NotImplementedError()

    def recommend_from_pool(self, *, pool: torch.Tensor, **kwargs):
        raise NotImplementedError()

    def _kwargs4full(self, kwargs: Dict):
        parameters = inspect.signature(self.recommend_from_full).parameters
        if any(param.kind == param.VAR_KEYWORD for param in parameters.values()):
            return kwargs
        return {key: kwargs[key] for key in parameters}

    def _kwargs4pool(self, kwargs: Dict):
        parameters = inspect.signature(self.recommend_from_pool).parameters
        if any(param.kind == param.VAR_KEYWORD for param in parameters.values()):
            return kwargs
        return {key: kwargs[key] for key in parameters}

    def recommend(self, **kwargs):
        if kwargs.get('pool', None) is None:
            return self.recommend_from_full(**self._kwargs4full(kwargs))
        else:
            return self.recommend_from_pool(**self._kwargs4pool(kwargs))

    @overload
    def encode(self, users: torch.Tensor) -> torch.Tensor:
        r"""
        User encoding.

        users: torch.Tensor, (B, 1)
        """

    @overload
    def encode(self, seqs: torch.Tensor) -> torch.Tensor:
        r"""
        Seq encoding.

        seqs: torch.Tensor, (B, S)
        """

    @overload
    def predict(self, users: torch.Tensor) -> torch.Tensor:
        r"""
        User-Item scoring.

        users: torch.Tensor, (B, 1)
        """

    @overload
    def predict(self, users: torch.Tensor, items: torch.Tensor) -> torch.Tensor:
        r"""
        User-Item scoring.

        users: torch.Tensor, (B, 1)
        items: torch.Tensor, (B, K + 1)
        """

    @overload
    def predict(self, users: torch.Tensor, positives: torch.Tensor, negatives: torch.Tensor) -> Tuple[torch.Tensor]:
        r"""
        User-Item scoring.

        users: torch.Tensor, (B, 1)
        positives: torch.Tensor, (B, 1)
        negatives: torch.Tensor, (B, K)
        """

    def predict(self, *args, **kwargs):
        raise NotImplementedError()

    def forward(self, *args, **kwargs):
        if self.training:
            return self.predict(*args, **kwargs)
        else:
            return self.recommend(**kwargs)




from typing import Iterable, Sequence

import random
import torchdata.datapipes as dp

from .base import BaseProcessor
from ..fields import SparseField


__all__ = [
    'RandomChoicedSource', 'RandomShuffledSource', 'OrderedSource',
    'RandomIDs', 'OrderedIDs', 'DummySource'
]


class RandomChoicedSource(BaseProcessor):
    r"""
    DataPipe that generates random items from given source.
    Note that this sampling is with replacement.

    Parameters:
    -----------
    source: Iterable 
        The source data to start.
    datasize: int 
        Datasize.
    """

    def __init__(
        self, source: Iterable, datasize: int
    ) -> None:
        super().__init__(None)

        self.source = tuple(source)
        self.datasize = datasize
        self._rng = random.Random()
        self.set_seed(1)

    def set_seed(self, seed: int):
        self._rng.seed(seed)

    def __iter__(self):
        for _ in range(self.datasize):
            yield self._rng.choice(self.source)


class RandomShuffledSource(BaseProcessor):
    r"""
    DataPipe that generates shuffled source.
    In this vein, every sample will be selected once per epoch.

    Parameters:
    -----------
    source: Iterable 
        The source data to start.
    """

    def __init__(self, source) -> None:
        super().__init__(None)

        self.source = list(source)
        self._rng = random.Random()
        self.set_seed(1)

    def set_seed(self, seed: int):
        self._rng.seed(seed)

    def __iter__(self):
        self._rng.shuffle(self.source)
        yield from iter(self.source)


class OrderedSource(BaseProcessor):
    r"""
    DataPipe that generates ordered items from given source.

    Parameters:
    -----------
    source: Sequence 
        The source data to start.
    datasize: int 
        Datasize.
    """

    def __init__(self, source: Sequence) -> None:
        super().__init__(None)

        assert isinstance(source, Sequence), f"Sequence type is required but received {type(source)} type."
        self.source = tuple(source)

    def __iter__(self):
        yield from iter(self.source)


class RandomIDs(RandomChoicedSource):
    r"""
    DataPipe that generates random IDs according to SparseField.

    Parameters:
    -----------
    field: SparseField 
        ID values to select from.
    datasize: int 
        Number of IDs to generate.
    """

    def __init__(
        self, field: SparseField,
        datasize: int,
    ) -> None:
        super().__init__(field.enums, datasize)


class OrderedIDs(OrderedSource):
    r"""
    DataPipe that generates ordered IDs.

    Parameters:
    -----------
    field: SparseField 
        ID values to select from.
    """

    def __init__(self, field: SparseField) -> None:
        super().__init__(field.enums)


class DummySource(OrderedSource):
    r"""
    DataPipe that generates dummy data.

    Parameters:
    -----------
    datasize: int 
        Number of data to generate.
    """
    def __init__(self, datasize: int) -> None:
        super().__init__(range(datasize))


@dp.functional_datapipe("dummy_")
class _DummySource(DummySource):
    r"""
    Functional DataPipe wrapper for DummySource.

    Parameters:
    -----------
    source_dp: dp.iter.IterDataPipe 
        The source DataPipe (No use here).
    datasize: int 
        Number of data to generate.
    """
    def __init__(self, source_dp: dp.iter.IterDataPipe, datasize: int) -> None:
        super().__init__(datasize)




"""
Refer to
    https://zhuanlan.zhihu.com/p/67287992
for formal definitions.
The implementations below are fully due to torchmetrics.
"""

from typing import Optional, Union, List 
from functools import  partial

import torch


__all__ = [
    'mean_abs_error', 'mean_squared_error', 'root_mse',
    'precision', 'recall', 'f1_score', 'auroc', 'hit_rate',
    'normalized_dcg', 'mean_reciprocal_rank', 'mean_average_precision'
]


def _reduce(reduction='mean'):
    r"""
    A decorator that applies a reduction operation to the output of a given function.

    Parameters:
    -----------
    reduction : str, optional
        The reduction operation to be applied to the output of the function. Should be one of {'none', 'mean', 'sum'}. Default is 'mean'.

    Returns:
    --------
    wrapper : function
        The decorated function.

    Notes:
    ------
    This decorator takes a function that takes two inputs: `preds` and `targets`, 
    both of which are either lists of PyTorch tensors or PyTorch tensors themselves. 
    The output of the function should also be a PyTorch tensor or a list of PyTorch tensors. 
    The decorator then applies a reduction operation specified by the `reduction` parameter to the output of the function.

    Examples:
    ---------
    Here is an example usage of this decorator:

    >>> @_reduce(reduction='mean')
    ... def mean_squared_error(preds, targets):
    ...     return F.mse_loss(preds, targets)

    Here, we define a function `mean_squared_error` that computes the mean squared error between `preds` and `targets` 
    using PyTorch's built-in mean squared error loss function (`F.mse_loss`). 
    The decorator `_reduce` is applied to this function, with `reduction` set to 'mean'. 
    This means that the output of `mean_squared_error` will be the mean of the mean squared error loss values computed for each element in `preds` and `targets`.
    """
    def decorator(func):
        def wrapper(
            preds: Union[List[torch.Tensor], torch.Tensor], 
            targets: Union[List[torch.Tensor], torch.Tensor], 
            *,
            reduction: str = reduction, **kwargs
        ):
            func_ = partial(func, **kwargs)
            if isinstance(preds, List):
                results = torch.tensor(list(map(
                    func_, preds, targets
                )))
            else:
                results = func_(preds, targets)
            if reduction == 'none':
                return results
            elif reduction == 'mean':
                return results.mean()
            elif reduction == 'sum':
                return results.sum()
            else:
                raise ValueError(f"reduction should be 'none'|'mean'|'sum' but {reduction} is received ...")
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    return decorator


# quality of predictions ========================================================================

@_reduce('mean')
def mean_abs_error(preds: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    r"""
    Computes Mean Absolute Error (MAE).

    .. math::

        \phi(x_i, y_i) = \frac{1}{d} \|x_i - y_i\|_1.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.


    Returns:
    --------
    Computed MAE: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> mean_abs_error(preds, targets, reduction='none')
    tensor([0.5000, 0.5250])
    >>> mean_abs_error(preds, targets)
    tensor(0.5125)
    """
    preds, targets = preds.float(), targets.float()
    return (preds - targets).abs().mean(-1)

@_reduce('mean')
def mean_squared_error(preds: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    r"""
    Calculate mean squared error (MSE).

    .. math::

        \phi(x_i, y_i) = \frac{1}{d} \|x_i - y_i\|_2^2.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.

    Returns:
    --------
    Computed MSE: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> mean_squared_error(preds, targets, reduction='none')
    tensor([0.3450, 0.3475])
    >>> mean_squared_error(preds, targets)
    tensor(0.3462)
    """
    preds, targets = preds.float(), targets.float()
    return (preds - targets).pow(2).mean(-1)

@_reduce('mean')
def root_mse(preds: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    r"""Root mean squared error (RMSE).

    .. math::

        \phi(x_i, y_i) = \sqrt{\frac{1}{d} \|x_i - y_i\|_2^2}.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.

    Returns:
    --------
    Computed RMSE: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> root_mean_squared_error(preds, targets, reduction='none')
    tensor([0.5874, 0.5895])
    >>> root_mean_squared_error(preds, targets)
    tensor(0.5884)
    """
    preds, targets = preds.float(), targets.float()
    return (preds - targets).pow(2).mean(-1).sqrt()


# quality of the set of recommendations ========================================================================

@_reduce('mean')
def precision(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    Compute the precision of the top-K predicted items for a given query.

    The precision is defined as the intersection between the top-K predicted items and the target items divided by
    the size of the top-K set.

    .. math::

        \phi(x_i, y_i) = (|Topk(x_i) \cap Topk(y_i)|) / |Topk(x_i)|,
    
    where :math: `Topk(\cdot)` returns the Top-K (predicted or ideal) items of the query :math: `i`,
    and :math: `|\cdot|` measures the cardinality of the set.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed Precision: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> precision(preds, targets, k=3, reduction='none')
    tensor([0.3333, 0.6667])
    >>> precision(preds, targets, k=3)
    tensor(0.5000)
    """

    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))
    indices = preds.topk(k)[1]
    relevant = targets.gather(-1, indices).sum(-1)
    return relevant / k

@_reduce('mean')
def recall(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    Recall is a measure of how well a ranking algorithm retrieves relevant items for a given query. 
    It is defined as the intersection of the top-K predicted items with the top-K relevant items, 
    divided by the total number of relevant items.

    .. math::

        \phi(x_i, y_i) = (|Topk(x_i) \cap Topk(y_i)|) / |y_i|,

    where :math:`Topk(\cdot)` returns the Top-K (predicted or ideal) items of the query :math:`i`,
    and :math:`|\cdot|` measures the cardinality of the set.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed Recall: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> recall(preds, targets, k=3, reduction='none')
    tensor([0.5000, 1.0000])
    >>> recall(preds, targets, k=3)
    tensor(0.7500)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))
    
    indices = preds.topk(k)[1]
    relevant = targets.gather(-1, indices).sum(-1)
    total = targets.sum(-1)
    invalid = total == 0
    relevant[invalid] = 0
    relevant[~invalid] /= total[~invalid]
    return relevant

@_reduce('mean')
def f1_score(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    The F1 score is the harmonic mean of precision and recall, where precision is the number of true positive predictions
    divided by the total number of positive predictions, and recall is the number of true positive predictions divided by
    the total number of positive ground truth labels.

    The F1 score is computed as follows:

    .. math::

        \phi(x_i, y_i) = (2 * Precision@k(x_i, y_i) * Recall@k(x_i, y_i)) / (Precision@k(x_i, y_i) + Recall@k(x_i, y_i))

    where :math: `Topk(\cdot)` returns the Top-K (predicted or ideal) items of the query :math: `i`,
    and :math: `|\cdot|` measures the cardinality of the set.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed F1-score: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> f1_score(preds, targets, k=3, reduction='none')
    tensor([0.4000, 0.8000])
    >>> f1_score(preds, targets, k=3)
    tensor(0.6000)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))
    precision_k = precision(preds, targets, k=k, reduction='none')
    recall_k = recall(preds, targets, k=k, reduction='none')
    score = 2 * precision_k * recall_k
    part2 = precision_k + recall_k
    valid = part2 != 0
    score[valid] /= part2[valid]
    return score

@_reduce("mean") #TODO: This implementation has not been verified.
def auroc(preds: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
    r"""
    AUC (Area Under the ROC Curve) represents the area under the Receiver Operating Characteristic (ROC) curve, 
    which is a curve showing the performance of a binary classifier system.

    .. math::

        \frac{n - (|N_+| - 1) / 2 - (\sum_{v \in |N_+|} rank(v)) / |N_+|}{n - |N_+|}

    where :math: `n` and :math: `|N_+|` denote the number of items and positive items, respectively; :math: `rank(v)` returns the rank of the positive item :math: `v`.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.

    Returns:
    --------
    Computed AUC: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> auroc(preds, targets, reduction='none')
    tensor([0.5000, 0.5000])
    >>> auroc(preds, targets) 
    tensor(0.5000)
    """
    preds, targets = preds.float(), targets.float()
    preds, indices = torch.sort(preds, descending=True)
    targets = targets.gather(-1, indices)
    ranks = torch.arange(1, targets.size(1) + 1, device=targets.device)
    length = targets.sum(dim=-1)
    results = targets.size(1)
    results -= (ranks * targets).sum(dim=-1).div(length)
    results -= (length - 1).div(2)
    results /= (targets.size(1) - length)
    results[torch.isnan(results)] = 0.
    return results

@_reduce('mean')
def hit_rate(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    Compute the Hit Rate score between predictions and targets.

    .. math::

        \phi(x_i, y_i) = \mathbb{I}(|Topk(x_i) \cap Topk(y_i)| \not = 0).
    
    where :math: `Topk(\cdot)` returns the Top-K (predicted or ideal) items of the query :math: `i`,
    and :math: `|\cdot|` measures the cardinality of the set.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed HitRate: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> hit_rate(preds, targets, k=3, reduction='none')
    tensor([1., 1.])
    >>> hit_rate(preds, targets, k=3)
    tensor(1.)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))

    indices = preds.topk(k)[1]
    relevant = targets.gather(-1, indices).sum(-1)
    return (relevant > 0).float()


# quality of the list of recommendations

def _dcg(target: torch.Tensor) -> torch.Tensor:
    """Computes Discounted Cumulative Gain for input tensor."""
    denom = torch.log2(torch.arange(target.shape[-1], device=target.device) + 2.0)
    return (target / denom).sum(dim=-1)

@_reduce('mean')
def normalized_dcg(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    NDCG (Normalized Discounted Cumulative Gain).

    .. math::

        \phi(x_i, y_i) = \frac{_dcg(x_i)}{_dcg(y_i)}

    where :math: `_dcg(\cdot)` computes the Discounted Cumulative Gain of the input.
        
    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed NDCG: torch.Tensor, shape (1,) or (N,)
        
    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> normalized_dcg(preds, targets, k=3, reduction='none')
    tensor([0.6131, 0.6934])
    >>> normalized_dcg(preds, targets, k=3)
    tensor(0.6533)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))

    indices = preds.topk(k)[1]
    sorted_target = targets.gather(-1, indices)
    ideal_target = targets.topk(k)[0]

    dcg = _dcg(sorted_target)
    ideal_dcg = _dcg(ideal_target)

    # filter undefined scores
    invalid = ideal_dcg == 0
    dcg[invalid] = 0
    dcg[~invalid] /= ideal_dcg[~invalid]
    return dcg

def _single_reciprocal_rank(preds: torch.Tensor, targets: torch.Tensor):
    if not targets.sum():
        return 0.
    positions = torch.nonzero(targets).view(-1)
    res = 1.0 / (positions[0] + 1.0)
    return res

@_reduce('mean')
def mean_reciprocal_rank(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    The mean reciprocal rank (MRR) is defined as the average of 
    the reciprocal rank of the first correct item in the predicted ranking of items for each query.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed MRR: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------

    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> mean_reciprocal_rank(preds, targets, reduction='none')
    tensor([1.0000, 0.5000])
    >>> mean_reciprocal_rank(preds, targets)
    tensor(0.7500)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))
    
    preds, indices = preds.topk(k=k, dim=-1)
    targets = targets.gather(-1, indices)

    if preds.ndim == 2:
        return torch.tensor(list(map(_single_reciprocal_rank, preds, targets)))
    else:
        return _single_reciprocal_rank(preds, targets)

def _single_adverage_precision(preds: torch.Tensor, targets: torch.Tensor):
    if not targets.sum():
        return 0.
    positions = torch.arange(1, len(targets) + 1, device=targets.device, dtype=torch.float32)[targets > 0]
    res = torch.div((torch.arange(len(positions), device=positions.device, dtype=torch.float32) + 1), positions).mean()
    return res

@_reduce('mean')
def mean_average_precision(preds: torch.Tensor, targets: torch.Tensor, *, k: Optional[int] = None) -> torch.Tensor:
    r"""
    MAP calculates the mean average precision for a set of predictions and targets.

    Parameters:
    -----------
    preds: torch.Tensor, shape (N, d) or (d,)
        Predictions or a single prediction tensor.
    targets: torch.Tensor, shape (N, d) or (d,)
        Targets tensor in accordance with preds.
    reduction: str, optional
        Specifies the reduction to apply to the output:
        - 'mean': the mean value of the error is returned. Default.
        - 'sum': the sum of the error is returned.
        - 'none': no reduction will be applied and a tensor of shape (n,) will be returned.
    k: int, optional
        - `int': Top-(K=k).
        - `None': Top-(K=d)

    Returns:
    --------
    Computed MAP: torch.Tensor, shape (1,) or (N,)

    Examples:
    ---------
    >>> preds = torch.tensor([[0.2, 0.3, 0.5, 0.], [0.1, 0.3, 0.5, 0.2]])
    >>> targets = torch.tensor([[0, 0, 1, 1], [0, 1, 0, 1]])
    >>> mean_average_precision(preds, targets, reduction='none')
    tensor([0.7500, 0.5833])
    >>> mean_average_precision(preds, targets)
    tensor(0.6667)
    """
    preds, targets = preds.float(), targets.float()
    if k is None:
        k = preds.size(-1)
    else:
        k = min(k, preds.size(-1))
    
    preds, indices = preds.topk(k=k, dim=-1)
    targets = targets.gather(-1, indices)

    if preds.ndim == 2:
        return torch.tensor(list(map(_single_adverage_precision, preds, targets)))
    else:
        return _single_adverage_precision(preds, targets)


if __name__ == "__main__":
    import doctest
    doctest.testmod()





import argparse


def tune(args):
    # Grid search for hyper-parameters can be conducted by:
    #
    #    freerec tune NAME_OF_EXPERIMENT CONFIG.yaml
    #
    from .parser import CoreParser
    from .launcher import Adapter
    cfg = CoreParser()
    cfg.compile(args)

    tuner = Adapter()
    tuner.compile(cfg)
    tuner.fit()


def make(args):
    # Make dataset:
    #
    #    freerec make DATASET --root=../data
    #
    from .data.preprocessing import datasets
    from .data.tags import USER, SESSION, ITEM, TIMESTAMP

    converter = getattr(datasets, args.dataset)(
        root=args.root,
        filename=args.filename
    )

    star4pos = args.star4pos
    kcore4user = args.kcore4user
    kcore4item = args.kcore4item
    ratios = tuple(map(int, args.ratios.split(',')))
    days = args.days
    strict = not args.not_strict
    
    if args.datatype == 'gen' and args.by == 'ratio':
        fields = None if args.all else (USER.name, ITEM.name)
        converter.make_general_dataset(
            star4pos=star4pos,
            kcore4user=kcore4user,
            kcore4item=kcore4item,
            strict=strict,
            ratios=ratios,
            fields=fields
        )
    elif args.datatype == 'seq' and args.by == 'leave-one-out':
        fields = None if args.all else (USER.name, ITEM.name, TIMESTAMP.name)
        converter.make_sequential_dataset(
            star4pos=star4pos,
            kcore4user=kcore4user,
            kcore4item=kcore4item,
            strict=strict,
            fields=fields
        )
    elif args.datatype == 'seq' and args.by == 'ratio':
        fields = None if args.all else (USER.name, ITEM.name, TIMESTAMP.name)
        converter.make_sequential_dataset_by_ratio(
            star4pos=star4pos,
            kcore4user=kcore4user,
            kcore4item=kcore4item,
            strict=strict,
            ratios=ratios,
            fields=fields
        )
    elif args.datatype == 'sess' and args.by == 'day':
        fields = None if args.all else (SESSION.name, ITEM.name, TIMESTAMP.name)
        converter.make_session_dataset_by_day(
            star4pos=star4pos,
            kcore4user=kcore4user,
            kcore4item=kcore4item,
            strict=strict,
            days=days,
            fields=fields
        )
    elif args.datatype == 'sess' and args.by == 'ratio':
        fields = None if args.all else (SESSION.name, ITEM.name, TIMESTAMP.name)
        converter.make_session_dataset_by_ratio(
            star4pos=star4pos,
            kcore4user=kcore4user,
            kcore4item=kcore4item,
            strict=strict,
            ratios=ratios,
            fields=fields
        )
    else:
        raise ValueError(f"`{args.datatype}' type dataset cannot be made by `{args.by}'")

def main():
    parser = argparse.ArgumentParser("FreeRec")
    subparsers = parser.add_subparsers()


    tune_parser = subparsers.add_parser("tune")

    tune_parser.set_defaults(func=tune)
    tune_parser.add_argument("description", type=str, help="...")
    tune_parser.add_argument("config", type=str, help="config.yml")
    tune_parser.add_argument("--exclusive", action="store_true", default=False, help="one by one or one for all")

    tune_parser.add_argument("--root", type=str, default=None, help="data")
    tune_parser.add_argument("--dataset", type=str, default=None, help="useless if no need to automatically select a dataset")
    tune_parser.add_argument("--device", type=str, default=None, help="device")
    tune_parser.add_argument("--num-workers", type=int, default=None)

    tune_parser.add_argument("--resume", action="store_true", default=False, help="resume the search from the recent checkpoint")


    make_parser = subparsers.add_parser("make")
    make_parser.set_defaults(func=make)

    make_parser.add_argument("dataset", type=str, help="dataset name")
    make_parser.add_argument("--root", type=str, default=".", help="data")
    make_parser.add_argument("--filename", type=str, default=None, help="filename of Atomic files")

    make_parser.add_argument(
        "--datatype", type=str, choices=('gen', 'seq', 'sess'), default='gen', 
        help="gen: general; seq: sequential; sess: session"
    )
    make_parser.add_argument(
        "--by", type=str, choices=('ratio', 'leave-one-out', 'day'), default='ratio', 
        help="gen: ratio; seq: leave-one-out; ratio; sess: ratio, day"
    )

    make_parser.add_argument("--star4pos", type=int, default=0, help="select interactions with `Rating > star4pos'")
    make_parser.add_argument("--kcore4user", type=int, default=10, help="select kcore interactions according to User")
    make_parser.add_argument("--kcore4item", type=int, default=10, help="select kcore interactions according to Item")
    make_parser.add_argument("--not-strict", action="store_true", default=False, help="filter by kcore once if True")
    make_parser.add_argument("--ratios", type=str, default="8,1,1", help="the ratios of training|validation|test set")
    make_parser.add_argument("--days", type=int, default=7, help="the second last days for validation and last days for test")

    make_parser.add_argument("--all", action="store_true", default=False, help="reserve all fields if True")

    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    main()